# This module is strictly for building expression trees.
# No tree transforms or visitors allowed!!!

import vocabulary as vo

__uniq_ctr__ = 0
def generate_uniqid(prefix='_', suffix=''):
    global __uniq_ctr__
    x = '%s%05d%s' % (prefix, __uniq_ctr__, suffix)
    __uniq_ctr__ += 1
    return x


class Expression(object):
    '''This class is what brings the terms (in vocabulary.py) to life.
    
    This is a pure functional data structure: do not try to mutate it.
    For more information read about "Purely Functional" on Wikipedia.'''

    # TODO: allow parameterizing this with alternative vocabularies

    def __init__(self, vocabulary_definition, *args, **kwargs):
        # to discourage people from calling __init__
        # (they should be calling function(...) instead):
        assert(kwargs['magic'] == 42)
        self.__vocabulary_definition__ = vocabulary_definition
        self.__args__ = args
        if vocabulary_definition.is_terminal:
            self.__sexp__ = (vocabulary_definition,) + args
        else:
            self.__sexp__ = (vocabulary_definition,) + tuple([x.sexp for x in args])

    # These are the only accessor methods for Expression:
    # {
    @property
    def definition(self): return self.__vocabulary_definition__
    @property
    def args(self):       return self.__args__
    # This one allows pattern matching a la OCaml et al. if used with
    # the Splarnektity module:
    @property
    def sexp(self):       return self.__sexp__
    # }

    # Comparison operations:
    def __eq__(self, other): return function(vo.EQ, self, other)
    def __ne__(self, other): return function(vo.NE, self, other)
    def __gt__(self, other): return function(vo.GT, self, other)
    def __lt__(self, other): return function(vo.LT, self, other)
    def __ge__(self, other): return function(vo.GE, self, other)
    def __le__(self, other): return function(vo.LE, self, other)

    # Math operations:
    def __add__(self, other):  return function(vo.ADD,      self, other)
    def __radd__(self, other): return function(vo.ADD,      other, self)
    def __sub__(self, other):  return function(vo.SUBTRACT, self, other)
    def __rsub__(self, other): return function(vo.SUBTRACT, other, self)
    def __mul__(self, other):  return function(vo.MULTIPLY, self, other)
    def __rmul__(self, other): return function(vo.MULTIPLY, other, self)
    def __div__(self, other):  return function(vo.DIVIDE,   self, other)
    def __rdiv__(self, other): return function(vo.DIVIDE,   other, self)
    # TODO:
    #def __truediv__(self, other):   return function()
    #def __rtruediv__(self, other):  return function()
    #def __floordiv__(self, other):  return function()
    #def __rfloordiv__(self, other): return function()
    #def __mod__(self, other):       return function()
    #def __rmod__(self, other):      return function()
    #def __divmod__(self, other):    return function()
    #def __rdivmod__(self, other):   return function()
    def __pow__(self, other):  return function(vo.POWER, self, other)
    def __rpow__(self, other): return function(vo.POWER, other, self)

    # Bit operations:
    def __lshift__(self, other):  return function(vo.LEFT_SHIFT,  self, other)
    def __rlshift__(self, other): return function(vo.LEFT_SHIFT,  other, self)
    def __rshift__(self, other):  return function(vo.RIGHT_SHIFT, self, other)
    def __rrshift__(self, other): return function(vo.RIGHT_SHIFT, other, self)
    def __and__(self, other):     return function(vo.BINARY_AND,  self, other)
    def __rand__(self, other):    return function(vo.BINARY_AND,  other, self)
    def __xor__(self, other):     return function(vo.BINARY_XOR,  self, other)
    def __rxor__(self, other):    return function(vo.BINARY_XOR,  other, self)
    def __or__(self, other):      return function(vo.BINARY_OR,   self, other)
    def __ror__(self, other):     return function(vo.BINARY_OR,   other, self)

    # ...and still more operations:
    def __neg__(self):         return function(vo.NEGATE, self)
    # Python calls this whenever abs() is called:
    def __abs__(self):         return function(vo.ABSOLUTE_VALUE, self)

    def __invert__(self):      return function(vo.INVERT, self) # ~x

# TODO: needs garbage collection and/or scope limiting:
__function_cache__ = { }
def function(vocabulary_definition, *args):
    if not vocabulary_definition.is_terminal:
        args = tuple(map(to_expression, args))
        key = (vocabulary_definition,) + args
    else:
        assert(len(args) == 1)
        # Note: here we key on type because (1,) and (1.0,) are
        #       equivalent keys when used to key a dict...
        #       and this is NOT what I want (it messes up X**2 versus
        #       X**2.0).
        # TODO: write a unit/regression test for this...
        key = (vocabulary_definition, type(args[0])) + args
    if key in __function_cache__:
        x = __function_cache__[key]
    else:
        x = Expression(vocabulary_definition, magic=42, *args)
        __function_cache__[key] = x
    return x

# Expression factories for other vocabulary definitions
# {
def var(name=None):
    if name is None: name = generate_uniqid('_cvexp_var_')
    return function(vo.VAR, name)

def const(value):
    return function(vo.CONST, value)

def logical_and(expr1, expr2):
    return function(vo.LOGICAL_AND, expr1, expr2)

def logical_or(expr1, expr2):
    return function(vo.LOGICAL_OR, expr1, expr2)

def abs(expression):
    return function(vo.ABSOLUTE_VALUE, expression)

def sin(expression):
    return function(vo.SIN, expression)

def cos(expression):
    return function(vo.COS, expression)

def integer(var):
    return function(vo.INTEGER, var)

def minimize(expression):
    return function(vo.MINIMIZE, expression)

def maximize(expression):
    return function(vo.MAXIMIZE, expression)
# }

def to_expression(x):
    '''Suggested usage:

        from cvexp.builders import to_expression as ex'''
    if   isinstance(x, Expression):               return x
    elif isinstance(x, (int, float, long, bool)): return const(x)
    elif isinstance(x, str):                      return var(x)
    else: raise Exception("Don't know how to convert to Expression: %s" %
                          repr(x)
                         )

# TODO:
#epsilon = Expression(None, magic=42)

