# Tested against CVXOPT 1.0

import _translate
import cvxopt.base
import cvxopt.solvers

def solve(exprs):
    '''Solves mixed integer linear, and quadratic programming problems.'''

    prob = _translate.OptimizationNormalForm(*exprs)

    def dense_matrix(rows, columns):
        return cvxopt.base.matrix(0.0, (rows, columns))

    def sparse_matrix(rows, columns):
        return cvxopt.base.spmatrix(0.0, [], [], (rows, columns))

    matrix = dense_matrix
    #matrix = sparse_matrix # TODO: make this to work

    # Compute some useful indexes:
    var_i = dict([(v, i) for i, v in enumerate(prob.vars)])
    i_var = dict([(i, v) for i, v in enumerate(prob.vars)])

    # Minimize:     (1/2) x^T P x + q^T x
    # Subject to:   G x <= h
    #               A x = b

    # XXX: DRY (with lpsolve55):
    def update_matrix_vector(M, v, row, poly):
        assert(poly.degree == 1)
        for term in poly.terms:
            if term.degree == 0:
                v[row, 0] = -term.coefficient
            else:
                [(var, exp)] = term.exponents.items()
                col = var_i[var]
                M[row, col] = term.coefficient
                # Make sure this is really a first order polynomial:
                assert(len(term.exponents) == 1)
                assert(exp == 1)

    # G x <= h constraints:
    G = matrix(len(prob.le), len(prob.vars))
    h = matrix(len(prob.le), 1)
    for ci, constraint in enumerate(prob.le):
        update_matrix_vector(G, h, ci, constraint)

    # A x = b constraints:
    A = matrix(len(prob.eq), len(prob.vars))
    b = matrix(len(prob.eq), 1)
    for ci, constraint in enumerate(prob.eq):
        update_matrix_vector(A, b, ci, constraint)

    assert(len(prob.int_vars) == 0) # TODO: not yet implemented...

    # (1/2) x^T P x + q^T x objective function:
    assert(len(prob.minimize) == 1) # don't know how to satisfy more than 1
    minimize_poly = prob.minimize[0]
    P = matrix(len(prob.vars), len(prob.vars))
    q = matrix(len(prob.vars), 1)
    for term in minimize_poly.terms:
        if   term.degree == 2:
            vis = [var_i[v] for v in term.exponents.keys()]
            if len(term.exponents) == 2:
                vi1, vi2 = vis
                P[vi1, vi2] += term.coefficient
                P[vi2, vi1] += term.coefficient
            elif len(term.exponents) == 1:
                vi = vis[0]
                # XXX: am I really supposed to do this twice?
                P[vi, vi] += term.coefficient
                P[vi, vi] += term.coefficient
            else:
                assert(False)
        elif term.degree == 1:
            assert(len(term.exponents) == 1)
            vi = var_i[term.exponents.keys()[0]]
            q[vi, 0] += term.coefficient
        else:
            assert(term.degree == 0)

    sol = cvxopt.solvers.qp(P, q, G, h, A, b)
    if sol['status'] == 'optimal':
        x = sol['x'] # this is the x vector (i.e. the solutions)
        result = { }
        for v, i in var_i.iteritems():
            result[v] = x[i, 0]
        return result
    else:
        print 'WARNING: cvxopt qp status:', qp['status']
        return None

def test():
    from builders import var
    from builders import minimize

    # TODO: make sure that the P matrix is right (especially the diagonals).
    #       do this by trying to balance out minimizations of x^2 + x*y
    # TODO: try out maximize
    # TODO: try out integer constraints

    def fuzzy_equal(a, b, fuzz):
        return (a - fuzz) < b < (a + fuzz)

    X = var('X')
    Y = var('Y')

    if True:
        sol = solve((
                    Y >= X,
                    Y >= 10 - X,
                    minimize(Y),
                   ))
        print 'X =', sol[X]
        print 'Y =', sol[Y]
        assert(fuzzy_equal(sol[X], 5, 0.000001))
        assert(fuzzy_equal(sol[Y], 5, 0.000001))

    if True:
        sol = solve((
                    minimize((X - 5) ** 2 + (Y - 3) ** 2),
                   ))
        print 'X =', sol[X]
        print 'Y =', sol[Y]
        assert(fuzzy_equal(sol[X], 5, 0.000001))
        assert(fuzzy_equal(sol[Y], 3, 0.000001))

    if True:
        sol = solve((
                    Y >= X,
                    Y >= 10 - X,
                    minimize(X * X + Y * Y),
                   ))
        print 'X =', sol[X]
        print 'Y =', sol[Y]
        assert(fuzzy_equal(sol[X], 5, 0.003)) # XXX: why so much error?
        assert(fuzzy_equal(sol[Y], 5, 0.003)) # XXX: why so much error?

if __name__ == '__main__':
    test()

