import _translate
import glpk
from glpk import LPX_MIN
from glpk import LPX_UP # upper bound
from glpk import LPX_LO # lower bound
from glpk import LPX_FX # equal bound
from glpk import LPX_FR # free bound (unbounded)
from glpk import LPX_IV # integer value constraint (v. LPX_CV)
from glpk import LPX_E_OK
from glpk import lpx_add_rows
from glpk import lpx_add_cols
from glpk import lpx_set_row_bnds
from glpk import lpx_set_col_bnds
from glpk import intArray
from glpk import doubleArray
from glpk import lpx_load_matrix
from glpk import lpx_set_col_kind
from glpk import lpx_set_obj_dir
from glpk import lpx_set_obj_coef
from glpk import lpx_write_cpxlp
from glpk import lpx_simplex
from glpk import lpx_intopt
from glpk import lpx_get_col_prim
from glpk import lpx_mip_col_val

def solve(exprs, out_filename=None):
    prob = _translate.OptimizationNormalForm(*exprs)

    lpx = glpk.lpx()
    lp = lpx.lp

    var_i = dict([(v, i) for i, v in enumerate(prob.vars)])
    i_var = dict([(i, v) for i, v in enumerate(prob.vars)])

    lpx_add_rows(lp, len(prob.le) + len(prob.eq))
    lpx_add_cols(lp, len(prob.vars))

    sparse_rows = []
    sparse_cols = []
    sparse_vals = []

    # XXX: DRY (with cvxopt):
    def extract_coefficients(poly):
        assert(poly.degree == 1)
        constant = 0
        coeffs = [0] * len(var_i)
        for term in poly.terms:
            if term.degree == 0:
                constant = term.coefficient
            else:
                [(var, exp)] = term.exponents.items()
                col = var_i[var]
                coeffs[col] = term.coefficient
                # Make sure this is really a first order polynomial:
                assert(len(term.exponents) == 1)
                assert(exp == 1)
        return constant, coeffs

    # XXX: is it really necessary to use float(...) below?

    # G x <= h constraints:
    coffset = 1
    for ci, constraint in enumerate(prob.le):
        row = coffset + ci
        # TODO: coeffs is not a sparse encoding:
        constant, coeffs = extract_coefficients(constraint)
        lpx_set_row_bnds(lp, row, LPX_UP, 0.0, -float(constant))
        for vi, coeff in enumerate(coeffs):
            if coeff != 0:
                col = vi + 1
                sparse_rows.append(row)
                sparse_cols.append(col)
                sparse_vals.append(float(coeff))

    # A x = b constraints:
    coffset = 1 + len(prob.le)
    for ci, constraint in enumerate(prob.eq):
        row = coffset + ci
        # TODO: coeffs is not a sparse encoding:
        constant, coeffs = extract_coefficients(constraint)
        lpx_set_row_bnds(lp, row, LPX_FX, -float(constant), -float(constant))
        for vi, coeff in enumerate(coeffs):
            if coeff != 0:
                col = vi + 1
                sparse_rows.append(row)
                sparse_cols.append(col)
                sparse_vals.append(float(coeff))

    # Unbound the columns:
    for v, iv in var_i.iteritems():
        col = iv + 1
        lpx_set_col_bnds(lp, col, LPX_FR, 0.0, 0.0)
        #lpx_set_col_bnds(lp, col, LPX_LO, 0.0, 0.0)

    sparse_rows1 = intArray(len(sparse_rows) + 1) # XXX: why +1??
    sparse_cols1 = intArray(len(sparse_cols) + 1) # XXX: why +1??
    sparse_vals1 = doubleArray(len(sparse_vals) + 1) # XXX: why +1??
    for i, x in enumerate(sparse_rows): sparse_rows1[i+1] = x
    for i, x in enumerate(sparse_cols): sparse_cols1[i+1] = x
    for i, x in enumerate(sparse_vals): sparse_vals1[i+1] = x
    lpx_load_matrix(lp, len(sparse_vals),
                    sparse_rows1, sparse_cols1, sparse_vals1)

    # TODO: integer constraints:
    #assert(len(prob.int_vars) == 0) # not yet implemented...
    for v in prob.int_vars:
        col = var_i[v] + 1 # column numbers start at 1
        lpx_set_col_kind(lp, col, LPX_IV)

    # q^T x objective function:
    assert(len(prob.minimize) == 1) # don't know how to satisfy more than 1
    lpx_set_obj_dir(lp, LPX_MIN)
    constant, min_coeffs = extract_coefficients(prob.minimize[0])
    for vi, coeff in enumerate(min_coeffs):
        if coeff != 0:
            col = vi + 1
            lpx_set_obj_coef(lp, col, float(coeff))

    if out_filename:
        lpx_write_cpxlp(lp, out_filename)

    mip = len(prob.int_vars) != 0

    if not mip:
        ret = lpx_simplex(lp) # simplex method (LP)

    #if len(prob.int_vars) and ret == LPX_E_OK:
    if mip:
        # NOTE: The intopt routine still requires the relaxed solution
        #       obtained from lpx_simplex ... I think ...
        #       ... wait, not it doesn't ...?
        ret = lpx_intopt(lp) # ??? method (MILP)
        # WARNING: result is not guaranteed to be optimal, even if ret is 0
        #          (but it should meet the constraints)
        # TODO: Check lpx_mip_status for LPX_I_OPT

    if ret == LPX_E_OK:
        result = { }
        for v, iv in var_i.iteritems():
            col = iv + 1
            if not mip: x = lpx_get_col_prim(lp, col)
            else:       x = lpx_mip_col_val(lp, col)
            if v in prob.int_vars:
                result[v] = int(x)
            else:
                result[v] = x
        return result
    else:
        return None

def test():
    from builders import var
    from builders import minimize
    from builders import integer

    # TODO: make sure that the P matrix is right (especially the diagonals).
    #       do this by trying to balance out minimizations of x^2 + x*y
    # TODO: try out maximize
    # TODO: try out integer constraints

    def fuzzy_equal(a, b, fuzz):
        return (a - fuzz) < b < (a + fuzz)

    X = var('X')
    Y = var('Y')

    if True:
        sol = solve((
                    Y == X,
                    Y >= 10 - X,
                    minimize(Y),
                   ), out_filename='glpk001.glpk')
        print 'X =', sol[X]
        print 'Y =', sol[Y]
        assert(fuzzy_equal(sol[X], 5, 0.000001))
        assert(fuzzy_equal(sol[Y], 5, 0.000001))

    if True:
        sol = solve((
                    Y == X,
                    Y >= 9.8 - X,
                    minimize(Y),
                   ), out_filename='glpk002.glpk')
        print 'X =', sol[X]
        print 'Y =', sol[Y]
        assert(fuzzy_equal(sol[X], 4.9, 0.000001))
        assert(fuzzy_equal(sol[Y], 4.9, 0.000001))

    if True:
        sol = solve((
                    Y + 0.1 == X,
                    Y >= 9.8 - X,
                    integer(Y),
                    minimize(Y),
                   ), out_filename='glpk003.glpk')
        print 'X =', sol[X]
        print 'Y =', sol[Y]
        assert(fuzzy_equal(sol[X], 5.1, 0.000001))
        assert(fuzzy_equal(sol[Y], 5.0, 0.000001))

if __name__ == '__main__':
    test()
