#!/usr/bin/env python

metadata = {
    'name': 'cvexp',
    'version': '0.1',
    'description':
      'Expression Tree Builder and Translator based on a Controlled Vocabulary',
    'author': 'David Baird',
    'author_email': 'dhbaird@gmail.com',
    'url': 'http://www.aclevername.com/projects/cvexp',
    'packages': ['cvexp'],
    'long_description': \
'''CVExp is short for "controlled vocabulary expressions."  CVExp is a
Python module that allows abstract syntax trees ("expressions") to be
constructed just as easily as writing normal Python code.  The names
of functions in the abstract syntax trees are based on a controlled
vocabulary, using URIs as names just like the W3C RDF does.

The intent of CVExp is that it should make it easier for developers
to integrate 3rd party libraries into Python without having to specify
entirely new APIs.  Libraries and languages of particular concern are
computer algebra systems, linear programming and optimization (such as
lp_solve, GLPK, and CVXOPT), and Prolog.

Here is an example usage to solve a mixed-integer linear programming
problem using GLPK as the backend.  In order for this to work, you
obviously must install prerequisite Python packages (i.e.  python-glpk or
whatever) because these are not included with CVExp::

    from cvexp.builders import var
    from cvexp.builders import integer
    from cvexp.builders import minimize
    from cvexp.translate_glpk import solve

    # You could also use lp_solve by doing this:
    # from cvexp.translate_lpsolve55 import solve
    # ...or you could use CVXOPT like this:
    # from cvexp.translate_cvxopt import solve

    X = var('X') # the 'X' name is optional
    Y = var('Y') # ...and so is 'Y'

    # Purely linear programming:
    sol = solve((
                 Y + 0.1 == X,
                 Y >= 9.8 - X,
                 minimize(Y),
               ), out_filename='problem.out') # out_filename is optional
    print 'X =', sol[X] # >>> 4.95
    print 'Y =', sol[Y] # >>> 4.85

    # Mixed integer-linear programming:
    sol = solve((
                 Y + 0.1 == X,
                 Y >= 9.8 - X,
                 integer(Y),
                 minimize(Y),
               ), out_filename='problem.out') # out_filename is optional
    print 'X =', sol[X] # >>> 5.1
    print 'Y =', sol[Y] # >>> 5

If using CVXOPT, quadratic programming problems can also be solved.
For example::

    from cvexp.builders import var
    from cvexp.builders import integer
    from cvexp.builders import minimize
    from cvexp.translate_cvxopt import solve

    X = var()
    Y = var()

    sol = solve((
                minimize((X - 5) ** 2 + (Y - 3) ** 2),
               ))
    print 'X =', sol[X] # >>> 5.0
    print 'Y =', sol[Y] # >>> 3.0

''',
    'classifiers': [
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Financial and Insurance Industry",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    'keywords': 'controlled vocabulary linear programming optimization',
    #'license': 'BSD',
    # I suspect that 'install_requires' isn't the right predicate to
    # user  here:
    'install_requires': ['sympy'], # sympy required for _translation.py
}

if __name__ == '__main__':
    #from distutils.core import setup
    from setuptools import setup # this allows: python build.py bdist_egg
    # WARNING: "python setup.py sdist" does not work on AFS.
    #          (due to hardlinks).  The following hack was found on the
    #          distutils mailing list 2005-August:
    import os
    delattr(os, 'link') # remove the ability to do hardlinks (AFS hates them)
    setup(**metadata)
